import backoff
from networkx import within_inter_cluster
import openai
import pydantic_core
from openai import AsyncOpenAI, OpenAI
from apropos.src.core.lms.cache_init import cache
from apropos.src.core.lms.vendors.openai_like import OpenAIStandardProvider
from apropos.src.core.lms.vendors.json_structured_outputs.core import (
    add_json_instructions_to_messages,
    extract_pydantic_model_from_response_async,
    extract_pydantic_model_from_response_sync,
)
import instructor

BACKOFF_TOLERANCE = 15  # 20


class OpenAIAPIProvider(OpenAIStandardProvider):
    def __init__(self, multi_threaded=True):
        self.sync_client = OpenAI()
        self.async_client = AsyncOpenAI()
        self.multi_threaded = multi_threaded
        # super().__init__(self.sync_client, self.async_client, multi_threaded=multi_threaded)
        self.supports_response_model = True

    @backoff.on_exception(
        backoff.expo,
        (pydantic_core._pydantic_core.ValidationError,),
        max_tries=BACKOFF_TOLERANCE,
        logger=None,
        on_giveup=lambda e: print(e)
        if isinstance(e, pydantic_core._pydantic_core.ValidationError)
        else None,
    )
    @backoff.on_exception(
        backoff.expo,
        (openai.RateLimitError, pydantic_core._pydantic_core.ValidationError),
        max_tries=BACKOFF_TOLERANCE,
        logger=None,
        on_giveup=lambda e: print(e)
        if isinstance(e, pydantic_core._pydantic_core.ValidationError)
        else None,
    )
    async def async_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        if "o1" in model:
            message = f"<instructions>{messages[0]['content']}</instructions>\n\n<information>{messages[1]['content']}</information>"
            messages = [{"role": "user", "content": message}]
            temperature = 1
        reminder = False
        retries = 2
        succeeded = False
        while not succeeded and retries > 0:
            try:
                messages_with_json_formatting_instructions = (
                    add_json_instructions_to_messages(
                        messages, response_model, reminder
                    )
                )
                raw_text_api_response = await self.async_chat_completion(
                    messages_with_json_formatting_instructions,
                    model,
                    temperature,
                    max_tokens,
                )
                structured_api_response = (
                    await extract_pydantic_model_from_response_async(
                        raw_text_api_response, response_model
                    )
                )
                succeeded = True
            except Exception as e:
                print("Error in async_chat_completion_with_response_model: ", e)
                reminder = True
                temperature += 0.01
                retries -= 1
        return structured_api_response

    @backoff.on_exception(
        backoff.expo,
        (openai.RateLimitError, pydantic_core._pydantic_core.ValidationError),
        max_tries=BACKOFF_TOLERANCE,
        logger=None,
        on_giveup=lambda e: print(e)
        if isinstance(e, pydantic_core._pydantic_core.ValidationError)
        else None,
    )
    def sync_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        if "o1" in model:
            message = f"<instructions>{messages[0]['content']}</instructions>\n\n<information>{messages[1]['content']}</information>"
            messages = [{"role": "user", "content": message}]
            temperature = 1
        reminder = False
        retries = 2
        succeeded = False
        while not succeeded and retries > 0:
            try:
                messages_with_json_formatting_instructions = (
                    add_json_instructions_to_messages(
                        messages, response_model, reminder
                    )
                )
                raw_text_api_response = self.sync_chat_completion(
                    messages_with_json_formatting_instructions,
                    model,
                    temperature,
                    max_tokens,
                )
                structured_api_response = extract_pydantic_model_from_response_sync(
                    raw_text_api_response, response_model
                )
                succeeded = True
            except Exception as e:
                print("Error in sync_chat_completion_with_response_model: ", e)
                reminder = True
                temperature += 0.01
                retries -= 1
        return structured_api_response


if __name__ == "__main__":
    import asyncio

    messages = [
        {
            "role": "system",
            "content": "You are a helpful assistant that can answer questions about the capital of France.",
        },
        {"role": "user", "content": "What is the capital of France?"},
    ]
    response = asyncio.run(
        OpenAIAPIProvider().async_chat_completion(
            messages=messages,
            model="o1-preview",
            temperature=0.0,
            max_tokens=500,
        )
    )
    print(response)



# Benchmark games used in the paper
benchmark_games_v0 = [
    "RockPaperScissorBasic",
    "BanditTwoArmedHighLowFixed",
    "MessengerL1",
    "MessengerL2",
    "Hanoi3Disk",
    "Crafter",
    "MinedojoCreative0",
]


from . import recorded_settings, game_challenges
import pandas as pd
import numpy as np


def normalize_score(game_name, score):
    # Normalize score to [0, 1] for a given environment

    if game_name in recorded_settings.keys():
        return (score - recorded_settings[game_name]["min score"]) / (
            recorded_settings[game_name]["human score"]
            - recorded_settings[game_name]["min score"]
        )
    else:
        raise ValueError(
            "Environment `{}` does not have recorded settings.".format(game_name)
        )


def analyze_capabilities(score_dict):
    # Analyze capabilities of a model based on the metric dictionary.
    # The metric dictionary should have the following format:
    # {
    #     'game_name': score,
    #     ...
    # }
    # The output is a dictionary with the following format:
    # {
    #     'game_name': capability,
    #     ...
    # }

    challenges_df = pd.DataFrame(game_challenges)
    challenges_df = challenges_df.loc[:, list(score_dict.keys())] - 1

    normalized_scores = np.array(list(score_dict.values())).reshape(-1, 1)
    scores = (challenges_df.values @ normalized_scores).squeeze() / np.sum(
        challenges_df.values, axis=1
    )

    return dict(zip(challenges_df.index, scores.flatten()))



from dataclasses import dataclass
from typing import Dict, List, Tuple, Callable, Literal
from copy import deepcopy
import random
import optuna
from apropos.src.bench.base import Benchmark, Question
from apropos.src.core.optimizers.base import DAGOptimizer
from apropos.src.core.programs.dag import LM_DAG, DagRecord
from apropos.src.core.programs.prompt import Demonstration, PromptTemplate, Topic
from pydantic import BaseModel
import concurrent.futures
import time
import os


class PromptDelta(BaseModel):
    description: str
    message: Literal["system", "user"]
    subcomponent: Literal["premise", "objective", "constraints", "user"]
    topic_name: str
    instructions_fields_edits: Dict[str, str]  # field_name, field_value
    # TODO: LATER
    # topic_template_edits: Dict[str,str]#before, after
    # input fields later ....

    def _apply_to_topic(self, topic: Topic) -> Topic:
        # if len(self.topic_template_edits) > 0:
        #     for before, after in self.topic_template_edits.items():
        #         topic.topic_template = topic.topic_template.replace(
        #             before, after
        #         )
        if len(self.instructions_fields_edits) > 0:
            for field_name, field_value in self.instructions_fields_edits.items():
                topic.instructions_fields[field_name] = field_value
        return topic

    def to_dict(self):
        return {
            "description": self.description,
            "message": self.message,
            "subcomponent": self.subcomponent,
            "topic_name": self.topic_name,
            "instructions_fields_edits": self.instructions_fields_edits,
        }

    def check_if_overlapping(self, prompt_delta: "PromptDelta") -> bool:
        return bool(
            set(self.instructions_fields_edits.keys())
            & set(prompt_delta.instructions_fields_edits.keys())
        )

    def validate(self, prompt: PromptTemplate):
        if self.message == "system":
            topics = getattr(prompt.system, self.subcomponent)
        elif self.message == "user":
            topics = getattr(prompt.user, self.subcomponent)
        else:
            raise ValueError(f"Invalid message type: {self.message}")
        for topic in topics:
            if topic.topic_name not in self.topic_name:
                raise ValueError(
                    f"Topic {self.topic_name} is not in {self.subcomponent} of prompt {self.message}"
                )

    def apply(self, prompt: PromptTemplate) -> PromptTemplate:
        if self.description == "null":
            return prompt
        prompt = deepcopy(prompt)
        if self.message == "system":
            topic_to_change, index_to_change = next(
                (
                    (topic, i)
                    for i, topic in enumerate(getattr(prompt.system, self.subcomponent))
                    if topic.topic_name == self.topic_name
                ),
                (None, -1),
            )
            if topic_to_change is None:
                raise ValueError(
                    f"Topic {self.topic_name} not found in {self.subcomponent} of prompt {self.message}"
                )
            changed_topic = self._apply_to_topic(topic_to_change)
            getattr(prompt.system, self.subcomponent)[index_to_change] = changed_topic
        elif self.message == "user":
            topic_to_change, index_to_change = next(
                (
                    (topic, i)
                    for i, topic in enumerate(prompt.user.user)
                    if topic.topic_name == self.topic_name
                ),
                (None, -1),
            )
            if topic_to_change is None:
                raise ValueError(
                    f"Topic {self.topic_name} not found in user of prompt {self.message}"
                )
            changed_topic = self._apply_to_topic(topic_to_change)
            getattr(prompt.user, self.subcomponent)[index_to_change] = changed_topic
        else:
            raise ValueError(f"Invalid message type: {self.message}")
        return prompt


@dataclass
class SearchSpace:
    prompt_delta_variations_by_problem_by_node: Dict[str, Dict[str, List[PromptDelta]]]
    demos_by_problem_by_node: Dict[str, Dict[str, List[Demonstration]]]


class RandomSearch_Optimizer:
    def __init__(
        self,
        seed: int,
        n_optuna_trials: int,
        max_demos_per_node: int = 5,
        questions_for_val: List[Question] = None,
    ):
        self.seed = seed
        self.n_optuna_trials = n_optuna_trials
        self.scored_attempts = []
        random.seed(self.seed)
        self.max_demos_per_node = max_demos_per_node
        self.questions_for_val = questions_for_val

    def search(
        self, baseline_program: LM_DAG, search_space: SearchSpace
    ) -> Tuple[LM_DAG, List[Tuple[List[Tuple[float, int]], LM_DAG]]]:
        def get_random_intervention():
            chosen_prompt_deltas_by_node = {
                node_name: [] for node_name in baseline_program.nodes.keys()
            }
            chosen_demos_by_node = {
                node_name: [] for node_name in baseline_program.nodes.keys()
            }
            for (
                problem_name,
                involved_node_names_with_deltas,
            ) in search_space.prompt_delta_variations_by_problem_by_node.items():
                for (
                    node_name,
                    prompt_delta_variations,
                ) in involved_node_names_with_deltas.items():
                    chosen_prompt_deltas_by_node[node_name].append(
                        random.choice(prompt_delta_variations)
                    )
            for problem_name, demos in search_space.demos_by_problem_by_node.items():
                for node_name, demos_for_node in demos.items():
                    chosen_demos_by_node[node_name].extend(
                        random.sample(demos_for_node, k=3)
                    )

            # Pare down any oversamples
            for node_name in chosen_demos_by_node.keys():
                chosen_demos_by_node[node_name] = random.sample(
                    chosen_demos_by_node[node_name],
                    min(len(chosen_demos_by_node[node_name]), self.max_demos_per_node),
                )
            return chosen_prompt_deltas_by_node, chosen_demos_by_node

        def to_candidate(chosen_prompt_deltas_by_node, chosen_demos_by_node):
            candidate_dag = deepcopy(baseline_program)
            for node_name, prompt_deltas in chosen_prompt_deltas_by_node.items():
                for prompt_delta in prompt_deltas:
                    candidate_dag.nodes[
                        node_name
                    ].transform.prompt = prompt_delta.apply(
                        candidate_dag.nodes[node_name].transform.prompt
                    )
            for node_name, demos in chosen_demos_by_node.items():
                candidate_dag.nodes[node_name].demonstrations = demos
            return candidate_dag

        best_score = float("-inf")
        best_params = None

        for _ in range(self.n_optuna_trials):
            chosen_prompt_deltas_by_node, chosen_demos_by_node = (
                get_random_intervention()
            )
            candidate_dag = to_candidate(
                chosen_prompt_deltas_by_node, chosen_demos_by_node
            )
            scores_with_records = [
                question.compute_and_score_attempt_sync(candidate_dag)
                for question in self.questions_for_val
            ]
            score = sum([score for score, _ in scores_with_records]) / len(
                scores_with_records
            )
            self.scored_attempts.append((score, scores_with_records, candidate_dag))

            if score > best_score:
                best_score = score
                best_params = {
                    "prompt_deltas": chosen_prompt_deltas_by_node,
                    "demos": chosen_demos_by_node,
                }
        best_candidate_dag = to_candidate(
            best_params["prompt_deltas"], best_params["demos"]
        )
        return best_candidate_dag, self.scored_attempts


class TPE_Optimizer:
    def __init__(
        self, seed: int, n_optuna_trials: int, questions_for_val: List[Question] = [], max_demos_per_node: int = 5
    ):
        self.seed = seed
        self.n_optuna_trials = n_optuna_trials
        self.questions_for_val = questions_for_val
        self.max_demos_per_node = max_demos_per_node
        pass

    def search(
        self, baseline_program: LM_DAG, search_space: SearchSpace
    ) -> Tuple[LM_DAG, List[Tuple[List[Tuple[float, int]], LM_DAG]]]:
        def get_intervention_for_trial(trial):
            chosen_prompt_deltas_by_node = {
                node_name: [] for node_name in baseline_program.nodes.keys()
            }
            chosen_demos_by_node = {
                node_name: [] for node_name in baseline_program.nodes.keys()
            }
            for (
                problem_name,
                involved_node_names_with_deltas,
            ) in search_space.prompt_delta_variations_by_problem_by_node.items():
                for (
                    node_name,
                    prompt_delta_variations,
                ) in involved_node_names_with_deltas.items():
                    chosen_prompt_deltas_by_node[node_name].append(
                        prompt_delta_variations[
                            trial.suggest_categorical(
                                f"prompt_delta_{node_name}_{problem_name}",
                                range(len(prompt_delta_variations)),
                            )
                        ]
                    )

            for (
                problem_name,
                demos_by_node,
            ) in search_space.demos_by_problem_by_node.items():
                for node_name, demos in demos_by_node.items():
                    num_demos = min(self.max_demos_per_node, len(demos))
                    for i in range(num_demos):
                        chosen_demo_index = trial.suggest_categorical(
                            f"demo_{node_name}_{problem_name}_{i}", range(len(demos))
                        )
                        chosen_demos_by_node[node_name].append(demos[chosen_demo_index])

            return chosen_prompt_deltas_by_node, chosen_demos_by_node

        def to_candidate(
            chosen_prompt_deltas_by_node: Dict[str, List[PromptDelta]],
            chosen_demos_by_node: Dict[str, List[Demonstration]],
        ):
            # import pdb; pdb.set_trace()
            assert isinstance(chosen_prompt_deltas_by_node, dict)
            assert all(
                isinstance(prompt_deltas, list)
                for prompt_deltas in chosen_prompt_deltas_by_node.values()
            )
            assert all(
                all(
                    isinstance(prompt_delta, PromptDelta)
                    for prompt_delta in prompt_deltas
                )
                for prompt_deltas in chosen_prompt_deltas_by_node.values()
            )
            assert isinstance(chosen_demos_by_node, dict)
            assert all(
                isinstance(demos, list) for demos in chosen_demos_by_node.values()
            )
            assert all(
                isinstance(demo, Demonstration)
                for demos in chosen_demos_by_node.values()
                for demo in demos
            )
            candidate_dag = deepcopy(baseline_program)
            for node_name, prompt_deltas in chosen_prompt_deltas_by_node.items():
                for prompt_delta in prompt_deltas:
                    candidate_dag.nodes[
                        node_name
                    ].transform.prompt = prompt_delta.apply(
                        candidate_dag.nodes[node_name].transform.prompt
                    )
            for node_name, demos in chosen_demos_by_node.items():
                candidate_dag.nodes[node_name].demonstrations = demos
            return candidate_dag

        scored_attempts = []

        # each trial specifies a delta for each node, including a null delta
        def create_objective() -> Callable:
            nonlocal scored_attempts

            def objective(trial):
                chosen_prompt_deltas_by_node, chosen_demos_by_node = (
                    get_intervention_for_trial(trial)
                )
                candidate_dag = to_candidate(
                    chosen_prompt_deltas_by_node, chosen_demos_by_node
                )
                print(f"Running trial {trial}")
                t0 = time.time()
                # Set the dag to parallelizable
                for node in candidate_dag.nodes.values():
                    node.transform.llm_config["multi_threaded"] = True
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=max(1, os.cpu_count() - 2)
                ) as executor:
                    futures = [
                        executor.submit(
                            question.compute_and_score_attempt_sync, candidate_dag
                        )
                        for question in self.questions_for_val
                    ]
                    scores_with_records = [
                        future.result()
                        for future in concurrent.futures.as_completed(futures)
                    ]
                score = sum([score for score, _ in scores_with_records]) / len(
                    scores_with_records
                )
                print(
                    f"... got score {score} in {time.time() - t0:.2f} seconds"
                )  # Would love to know why it stalls / is so slow sometimes
                scored_attempts.append((score, scores_with_records, candidate_dag))
                return score

            return objective

        print("Beginning TPE search")
        objective = create_objective()
        sampler = optuna.samplers.TPESampler(seed=self.seed, multivariate=True)
        study = optuna.create_study(direction="maximize", sampler=sampler)
        study.optimize(objective, n_trials=self.n_optuna_trials)
        best_chosen_prompt_deltas_by_node, best_chosen_demos_by_node = (
            get_intervention_for_trial(study.best_trial)
        )
        best_candidate_dag = to_candidate(
            best_chosen_prompt_deltas_by_node, best_chosen_demos_by_node
        )
        scores = [score for score, _, _ in scored_attempts]
        print("Scores: ", scores)
        return best_candidate_dag, scored_attempts



import gym
from gym import error, spaces, utils
from gym.utils import seeding
from ..utils import HistoryTracker, describe_act

import random
import itertools
import numpy as np
import math
import cc3d
import minedojo

biomes_dict = {
    0: "Ocean",
    1: "Plains",
    2: "Desert",
    3: "Extreme Hills",
    4: "Forest",
    5: "Taiga",
    6: "Swampland",
    7: "River",
    8: "Hell (The Nether)",
    9: "The End",
    10: "FrozenOcean",
    11: "FrozenRiver",
    12: "Ice Plains",
    13: "Ice Mountains",
    14: "MushroomIsland",
    15: "MushroomIslandShore",
    16: "Beach",
    17: "DesertHills",
    18: "ForestHills",
    19: "TaigaHills",
    20: "Extreme Hills Edge",
    21: "Jungle",
    22: "JungleHills",
    23: "JungleEdge",
    24: "Deep Ocean",
    25: "Stone Beach",
    26: "Cold Beach",
    27: "Birch Forest",
    28: "Birch Forest Hills",
    29: "Roofed Forest",
    30: "Cold Taiga",
    31: "Cold Taiga Hills",
    32: "Mega Taiga",
    33: "Mega Taiga Hills",
    34: "Extreme Hills+",
    35: "Savanna",
    36: "Savanna Plateau",
    37: "Mesa",
    38: "Mesa Plateau F",
    39: "Mesa Plateau",
    127: "The Void",
    129: "Sunflower Plains",
    130: "Desert M",
    131: "Extreme Hills M",
    132: "Flower Forest",
    133: "Taiga M",
    134: "Swampland M",
    140: "Ice Plains Spikes",
    149: "Jungle M",
    151: "JungleEdge M",
    155: "Birch Forest M",
    156: "Birch Forest Hills M",
    157: "Roofed Forest M",
    158: "Cold Taiga M",
    160: "Mega Spruce Taiga",
    161: "Redwood Taiga Hills M",
    162: "Extreme Hills+ M",
    163: "Savanna M",
    164: "Savanna Plateau M",
    165: "Mesa (Bryce)",
    166: "Mesa Plateau F M",
    167: "Mesa Plateau M",
}

yaw_granularity = 6
pitch_granularity = 6
FOV = 96

pitch_cnt = len(np.arange(-FOV // 2, FOV // 2 + 1, pitch_granularity))
yaw_cnt = len(np.arange(-FOV // 2, FOV // 2 + 1, yaw_granularity))


def get_direction(yaw, pitch):
    # Convert yaw and pitch to radians
    yaw_rad = math.radians(yaw)
    pitch_rad = math.radians(pitch)
    direction_list = []

    # Calculate the x, y, and z components of the direction vector
    x = math.sin(yaw_rad) * math.cos(pitch_rad)
    y = math.sin(pitch_rad)
    z = math.cos(yaw_rad) * math.cos(pitch_rad)

    # Determine the direction based on the sign of the x, y, and z components
    if y > 0.2:
        direction = "above you"
    elif y < -0.2:
        direction = "below you"
    else:
        direction = "at your level"

    if z > 0.2:
        direction_list.append("north")
    elif z < -0.2:
        direction_list.append("south")

    if x > 0.2:
        direction_list.append("east")
    elif x < -0.2:
        direction_list.append("west")

    return (
        direction
        + " to "
        + ("front" if "-".join(direction_list) == "" else "-".join(direction_list))
    )


def describe_surround(obs):
    # print("What blocks are around me?")
    # print()
    result = "Around you:\n"

    # Voxel size (Can be changed based on the voxel space)
    voxel_x_size = 9
    voxel_y_size = 9
    voxel_z_size = 9
    voxel_center_x = voxel_x_size // 2
    voxel_center_y = voxel_y_size // 2
    voxel_center_z = voxel_z_size // 2

    # Convert the 3D block name list into np array
    block_names = np.array(obs["voxels"]["block_name"])

    # Only preserve blocks 1 meter below me
    block_names = block_names[:, voxel_z_size // 2 - 1 :, :]

    # Encode the block name using its unique index
    unique_block_names, unique_block_indices = np.unique(
        block_names, return_inverse=True
    )
    block_names_int = np.array(unique_block_indices).reshape(block_names.shape)

    # Apply 3D connected component and extract the number of labels
    block_names_labels, labels_count = cc3d.connected_components(
        block_names_int, return_N=True, connectivity=6
    )

    # Get the centroids of each label
    centroids = cc3d.statistics(block_names_labels)["centroids"]

    # Get the map: label -> block name
    label_2_block_name = [""] * (labels_count + 1)
    for i in range(voxel_x_size):
        for j in range(block_names.shape[1]):
            for k in range(voxel_y_size):
                label = block_names_labels[i][j][k]
                block_name = block_names[i][j][k]
                label_2_block_name[label] = block_name

    # Describe the surrounding environment based on the connected component labels
    # Helper lists for printing
    x_axis_list = ["west", "east"]
    y_axis_list = ["north", "south"]
    z_axis_list = [
        "below you",
        "above you",
    ]  # The height of me is defined as the height of my feet

    # The map: block name -> min distance
    block2dist = dict()

    # The map: block name -> direction
    block2dir = dict()

    # Get the direction of each label based on the centroid
    for i in range(labels_count + 1):
        middle_point = centroids[i]
        block_name = label_2_block_name[i]

        # Skip the air
        if block_name == "air":
            continue

        # Get the indices to the printing help list
        x_coord = 1 if middle_point[0] > voxel_center_x else 0
        z_coord = 1 if middle_point[1] > 1 else 0
        y_coord = 1 if middle_point[2] > voxel_center_y else 0

        is_x_center = math.isclose(middle_point[0], voxel_center_x)
        is_z_center = math.isclose(middle_point[1], 1)
        is_y_center = math.isclose(middle_point[2], voxel_center_y)

        # If the component is exactly centered, skip
        if is_x_center and is_y_center and is_z_center:
            continue

        # Update the min distance and correspoding direction for the same block name
        distance = math.sqrt(
            (middle_point[0] - voxel_center_x) ** 2
            + (middle_point[1] - 1) ** 2
            + (middle_point[2] - voxel_center_y) ** 2
        )
        if block_name not in block2dist or distance < block2dist[block_name]:
            block2dist[block_name] = distance

            # Record the direction of this component
            direction_str_list = []

            if not is_z_center:
                direction_str_list.append(z_axis_list[z_coord])
            else:
                direction_str_list.append("")

            if not is_y_center:
                direction_str_list.append(y_axis_list[y_coord])

            # Skip when the coordinate is at the voxel center
            if not is_x_center:
                direction_str_list.append(x_axis_list[x_coord])

            direction_str = (
                direction_str_list[0]
                + " to "
                + (
                    "-".join(direction_str_list[1:])
                    if "-".join(direction_str_list[1:]) != ""
                    else "front"
                )
            )

            block2dir[block_name] = direction_str.strip()

    # Finally print the blocks
    for block_name in block2dist:
        direction_str = block2dir[block_name]
        result += f" - {block_name}, {'%.2f' % block2dist[block_name]} blocks away, {direction_str}\n"

    return result.strip()


# Describe the block the agent is facing to


def describe_cursor(obs):
    # Track the cursor using lidar
    cursor_block = obs["rays"]["block_name"][0]

    # Skip if the cursor is pointing to air
    if cursor_block == "air":
        return "You're not aiming at any block."
    else:
        return f"You're aiming at {cursor_block}."


# Describe the surrounding entities around the agent


def describe_entity(obs):
    result = ""
    my_yaw = obs["location_stats"]["yaw"]
    my_pitch = obs["location_stats"]["pitch"]

    # Number of pitch/yaw rays (Can be changed based on the lidar rays)

    # A flag indicating where there are entities around
    see_entity = False

    # Reshape and convert the list into np array
    entity_names = np.array(obs["rays"]["entity_name"][1:].reshape(pitch_cnt, yaw_cnt))

    # Encode the block name using its unique index
    unique_entity_names, unique_entity_indices = np.unique(
        entity_names, return_inverse=True
    )
    entity_names_int = np.array(unique_entity_indices).reshape(entity_names.shape)

    # Apply 2D connected components
    entity_names_labels, labels_count = cc3d.connected_components(
        entity_names_int, return_N=True, connectivity=8
    )

    # Get the map: label -> entity name
    label_2_entity_name = [""] * (labels_count + 1)
    for i in range(pitch_cnt):
        for j in range(yaw_cnt):
            label = entity_names_labels[i][j]
            entity_name = entity_names[i][j]
            label_2_entity_name[label] = entity_name

    # Describe each component
    for i in range(labels_count + 1):
        entity_name = label_2_entity_name[i]

        # Skip the null entity
        if entity_name == "null":
            continue
        else:
            see_entity = True

        # Find all the indices of this component
        all_idx = np.where(entity_names_labels == i)
        amount = np.sum(entity_names_labels == i)

        # Find the minimum distance inside each component
        min_distance = math.inf
        direction = ""
        for row, col in zip(all_idx[0], all_idx[1]):
            index = row * yaw_cnt + col
            distance = obs["rays"]["entity_distance"][index + 1]
            if distance < min_distance:
                min_distance = distance
                yaw = ((col / (yaw_cnt - 1)) * FOV - FOV / 2 + my_yaw + 540) % 360 - 180
                pitch = (row / (pitch_cnt - 1)) * 180 - 90 + my_pitch
                direction = get_direction(yaw, pitch)

        amount_description = "taking {0:.0f}% of screen".format(
            amount / (pitch_cnt * yaw_cnt) * 100
        )

        result += f" - {entity_name}, {'%.2f' % min_distance} blocks away, {direction}, {amount_description}\n"

    # If there are no surrounding entities that the agent can see
    if not see_entity:
        result = ""

    return result


# Describe the surrounding entities around the agent


def describe_obj(obs):
    result = "You see:\n"
    my_yaw = obs["location_stats"]["yaw"]
    my_pitch = obs["location_stats"]["pitch"]

    # Number of pitch/yaw rays (Can be changed based on the lidar rays)

    # A flag indicating where there are entities around
    see_entity = False

    # Reshape and convert the list into np array
    entity_names = np.array(obs["rays"]["block_name"][1:].reshape(pitch_cnt, yaw_cnt))

    # Encode the block name using its unique index
    unique_entity_names, unique_entity_indices = np.unique(
        entity_names, return_inverse=True
    )
    entity_names_int = np.array(unique_entity_indices).reshape(entity_names.shape)

    # Apply 2D connected components
    entity_names_labels, labels_count = cc3d.connected_components(
        entity_names_int, return_N=True, connectivity=8
    )

    # Get the map: label -> entity name
    label_2_entity_name = [""] * (labels_count + 1)
    for i in range(pitch_cnt):
        for j in range(yaw_cnt):
            label = entity_names_labels[i][j]
            entity_name = entity_names[i][j]
            label_2_entity_name[label] = entity_name

    # Describe each component
    for i in range(labels_count + 1):
        entity_name = label_2_entity_name[i]

        # Skip the null entity
        if entity_name in ("null", "air"):
            continue
        else:
            see_entity = True

        # Find all the indices of this component
        all_idx = np.where(entity_names_labels == i)
        amount = np.sum(entity_names_labels == i)

        # Find the minimum distance inside each component
        min_distance = math.inf
        direction = ""
        for row, col in zip(all_idx[0], all_idx[1]):
            index = row * yaw_cnt + col
            distance = obs["rays"]["block_distance"][index + 1]
            if distance < min_distance:
                min_distance = distance
                yaw = ((col / (yaw_cnt - 1)) * FOV - FOV / 2 + my_yaw + 540) % 360 - 180
                pitch = ((row / (pitch_cnt - 1)) * 180 - 90) + my_pitch
                direction = get_direction(yaw, pitch)

        amount_description = "taking {0:.0f}% of screen".format(
            amount / (pitch_cnt * yaw_cnt) * 100
        )

        result += f" - {entity_name}, {'%.2f' % min_distance} blocks away, {direction}, {amount_description}\n"

    # If there are no surrounding entities that the agent can see
    if not see_entity:
        result = ""

    return result.strip()


# Describe the exact coordinate the agent is in
def describe_location(obs):
    result = ""

    coord_list = obs["location_stats"]["pos"]

    # Describe the direction the agent is currently facing
    yaw = (obs["location_stats"]["yaw"] + 180) % 360 - 180

    direction_list = [
        "north",
        "north-east",
        "east",
        "south-east",
        "south",
        "south-west",
        "west",
        "north-west",
    ]
    direction_index = (
        int((yaw // 22.5) + 1)
    ) // 2  # Calculate the index mapping to the direction_list above
    if direction_index == 4:
        direction_index *= -1
    direction_index += 4

    result = f"Coordinate ({'%.2f' % coord_list[0]},{'%.2f' % coord_list[1]},{'%.2f' % coord_list[2]}). Facing {direction_list[direction_index]}."  # Describe whether the agent is looking up or down
    pitch = obs["location_stats"]["pitch"]

    if pitch < 0:
        result += " Looking up."
    elif pitch > 0:
        result += " Looking down."

    return result.strip()


def describe_frame(obs):
    return (
        describe_location(obs)
        + "\n\n"
        + describe_cursor(obs)
        + "\n"
        + describe_surround(obs)
        + "\n"
        + describe_entity(obs)
        + "\n"
        + describe_obj(obs)
    )


class MineDojoEnv(gym.Env):
    default_iter = 20
    default_steps = 200

    def __init__(self, task_id, max_steps=2):
        # Creative task
        self._env = minedojo.make(
            task_id="creative:{}".format(task_id),
            fast_reset=False,
            image_size=(160, 256),
            use_voxel=True,
            voxel_size=dict(
                xmin=-4, ymin=-4, zmin=-4, xmax=4, ymax=4, zmax=4
            ),  # Set voxel space to be 9*9*9
            use_lidar=True,
            lidar_rays=[(0, 0, 3)]
            + [
                (np.pi * pitch / 180, np.pi * yaw / 180, 65)
                for pitch in np.arange(-FOV // 2, FOV // 2 + 1, pitch_granularity)
                for yaw in np.arange(-FOV // 2, FOV // 2 + 1, yaw_granularity)
            ],  # Track the agent's cursor
        )
        goal = self._env.task_prompt.replace(".", "")[
            self._env.task_prompt.find("find ") + len("find ") :
        ]
        self.goal_set = {k for k, v in biomes_dict.items() if goal in v.lower()}
        self.action_list = ["Move North", "Move East", "Move South", "Move West"]
        self.noop_step = 50

        self.history = HistoryTracker(max_steps)
        self.desc = """
You are in Minecraft and your goal is to find a {} biome. You are not allowed to craft anything.

In your observation, you are provided the amount of space an object takes in your field of view. Note that objects of the same size takes more space when they are closer to you.

{}
""".format(goal, describe_act(self.action_list)).strip()

    def get_turn_action(self, act, obs):
        yaw = obs["location_stats"]["yaw"]
        # Compute yaw delta to face the yaw indicated by the action 'act'
        action_yaw = act * 90 - 180
        yaw_delta = (action_yaw - yaw + 180) % 360 // 15
        action = self._env.action_space.no_op()
        action[4] = yaw_delta
        return action

    def go_forward_action(self):
        action = self._env.action_space.no_op()
        action[0] = 1  # Move forward
        action[1] = (
            0 if random.random() < 0.5 else (1 if random.random() < 0.5 else 2)
        )  # Move left/right with prob=0.5 to get unstuck
        action[2] = 3 if random.random() < 0.5 else 1  # Sprint with prob=0.5
        return action

    def describe(self, obs, action=None):
        if action is not None:
            return "You took action {}.\n\n{}".format(
                self.action_list[action], describe_frame(obs)
            )
        else:
            return describe_frame(obs)

    def reset(self):
        obs = self._env.reset()
        self.history.reset()
        info = {
            "obs": self.describe(obs),
            "manual": self.desc,
            "history": self.history.describe(),
            "score": 0,
            "completed": 0,
        }
        self.history.step(info)
        self.obs = obs
        return obs, info

    def step(self, action):
        R = 0
        obs, reward, done, info = self._env.step(self.get_turn_action(action, self.obs))
        R += reward
        for _ in range(self.noop_step):
            obs, reward, done, info = self._env.step(self.go_forward_action())
            R += reward
            if obs["location_stats"]["biome_id"].item() in self.goal_set:
                break

        info.update(
            {
                "obs": self.describe(obs, action),
                "manual": self.desc,
                "history": self.history.describe(),
                "score": 0,
                "completed": 0,
            }
        )
        if obs["location_stats"]["biome_id"].item() in self.goal_set:
            info["completed"] = 1
            info["score"] = 1
            done = True
        self.history.step(info)
        self.obs = obs
        return obs, R, done, info



from typing import Dict, List, Type
from dataclasses import dataclass
import json

import loguru
import sqlite3
import json
from pydantic import BaseModel
from typing import Dict, List, Type
from dataclasses import dataclass
import json
import loguru
from diskcache import Cache
from pydantic import BaseModel
import hashlib
import time

logger = loguru.logger

DISKCACHE_SIZE_LIMIT = 10 * 1024 * 1024 * 1024  # 10 GB


@dataclass
class SafeCache:
    def __init__(self, fast_cache_dir, slow_cache_db):
        self.fast_cache = Cache(fast_cache_dir, size_limit=DISKCACHE_SIZE_LIMIT)
        self.conn = sqlite3.connect(slow_cache_db)
        self.cursor = self.conn.cursor()
        self.cursor.execute("""CREATE TABLE IF NOT EXISTS cache
                              (key TEXT PRIMARY KEY, response TEXT, information TEXT)""")
        self.conn.commit()

    def get_cache_key(
        self,
        messages: List[Dict],
        model: str,
        temperature: float,
        response_model: Type[BaseModel],
    ):
        if not all([isinstance(msg["content"], str) for msg in messages]):
            normalized_messages = "".join([str(msg["content"]) for msg in messages])
        else:
            normalized_messages = "".join([msg["content"] for msg in messages])
        normalized_model = model
        normalized_temperature = f"{temperature:.2f}"[:4]
        normalized_response_model = (
            str(response_model.schema()) if response_model else ""
        )
        return hashlib.sha256(
            (
                normalized_messages
                + normalized_model
                + normalized_temperature
                + normalized_response_model
            ).encode()
        ).hexdigest()

    def hit_cache(self, messages, model, temperature, response_model):
        fast_result = self.hit_cache_fast(messages, model, temperature, response_model)
        if fast_result and isinstance(fast_result, dict) and "response" in fast_result:
            fast_result = fast_result["response"]
        if fast_result and isinstance(fast_result, dict):
            return response_model(**fast_result)
        elif fast_result:
            return fast_result
        slow_result = self.hit_cache_slow(messages, model, temperature, response_model)
        if slow_result:
            return (
                slow_result
                if not "response" in slow_result
                else slow_result["response"]
            )
        return None

    def hit_cache_fast(self, messages, model, temperature, response_model):
        key = self.get_cache_key(messages, model, temperature, response_model)
        if key in self.fast_cache:
            try:
                cache_data = self.fast_cache[key]
            except AttributeError:
                return None
            if response_model is not None:
                if isinstance(cache_data["response"], dict):
                    response = cache_data["response"]
                    return response_model(**(response))
            return cache_data["response"]
        return None

    def hit_cache_slow(self, messages, model, temperature, response_model):
        key = self.get_cache_key(messages, model, temperature, response_model)
        self.cursor.execute("SELECT response FROM cache WHERE key = ?", (key,))
        result = self.cursor.fetchone()
        if result:
            return json.loads(result[0])
        return None

    def hit_cache_brute_force(self, messages, model, temperature, response_model):
        target_features = self.get_cached_features(
            messages, model, temperature, response_model
        )
        target_system = target_features["system"]
        target_user = target_features["user"]
        target_model = target_features["model"]

        self.cursor.execute("SELECT response, information FROM cache")
        results = self.cursor.fetchall()

        for response, information in results:
            info = json.loads(information)
            if (
                info["system"] == target_system
                and info["user"] == target_user
                and info["model"] == target_model
            ):
                return json.loads(response)
        return None

    def get_cached_features(
        self,
        messages: List[Dict],
        model: str,
        temperature: float,
        response_model: Type[BaseModel],
    ):
        if len(messages) == 2:
            return {
                "system": hashlib.sha256((messages[0]["content"]).encode()).hexdigest(),
                "user": hashlib.sha256(
                    (str(messages[1]["content"])).encode()
                ).hexdigest(),
                "model": model,
                "temperature": f"{temperature:.2f}"[:4],
            }
        elif len(messages) == 1:
            return {
                "system": hashlib.sha256("".encode()).hexdigest(),
                "user": hashlib.sha256((messages[0]["content"]).encode()).hexdigest(),
                "model": model,
                "temperature": f"{temperature:.2f}"[:4],
            }

    def add_to_cache(self, messages, model, temperature, response_model, response):
        key = self.get_cache_key(messages, model, temperature, response_model)
        information = json.dumps(
            self.get_cached_features(messages, model, temperature, response_model)
        )

        if isinstance(response, BaseModel):
            response_dict = response.model_dump()
            response_class = response.__class__.__name__
        else:
            response_dict = response
            response_class = None

        cache_data = {
            "response": response_dict,
            "response_class": response_class,
            "information": information,
        }

        self.fast_cache[key] = cache_data
        self.cursor.execute(
            "INSERT OR REPLACE INTO cache (key, response, information) VALUES (?, ?, ?)",
            (key, json.dumps(cache_data), information),
        )
        self.conn.commit()
        return key

    def dump_fast_to_slow(self):
        for key in self.fast_cache:
            cache_data = self.fast_cache[key]
            self.cursor.execute("SELECT key FROM cache WHERE key = ?", (key,))
            if not self.cursor.fetchone():
                self.cursor.execute(
                    "INSERT INTO cache (key, response, information) VALUES (?, ?, ?)",
                    (
                        key,
                        json.dumps(cache_data["response"]),
                        cache_data["information"],
                    ),
                )
        self.conn.commit()

    def close(self):
        self.dump_fast_to_slow()
        self.fast_cache.close()
        self.conn.close()


@dataclass
class ThreadedCache:
    def __init__(self, fast_cache_dir):
        self.fast_cache = Cache(fast_cache_dir, size_limit=DISKCACHE_SIZE_LIMIT)

    def hit_cache(self, messages, model, temperature, response_model):
        key = self.get_cache_key(messages, model, temperature, response_model)
        if key in self.fast_cache:
            try:
                cache_data = self.fast_cache[key]
            except AttributeError:
                return None
            if response_model is not None:
                if isinstance(cache_data["response"], dict):
                    response = cache_data["response"]
                    return response_model(**response)
            if isinstance(cache_data, str):
                cache_data = {
                    "response": cache_data,
                    "response_class": None,
                    "information": None,
                }
            return cache_data["response"]
        return None

    def add_to_cache(self, messages, model, temperature, response_model, response):
        key = self.get_cache_key(messages, model, temperature, response_model)
        information = json.dumps(
            self.get_cached_features(messages, model, temperature, response_model)
        )

        if isinstance(response, BaseModel):
            response_dict = response.model_dump()
            response_class = response.__class__.__name__
        else:
            response_dict = response
            response_class = None

        cache_data = {
            "response": response_dict,
            "response_class": response_class,
            "information": information,
        }

        self.fast_cache[key] = cache_data
        return key

    def get_cached_features(self, messages, model, temperature, response_model):
        if len(messages) == 2:
            return {
                "system": hashlib.sha256((messages[0]["content"]).encode()).hexdigest(),
                "user": hashlib.sha256(
                    (str(messages[1]["content"])).encode()
                ).hexdigest(),
                "model": model,
                "temperature": f"{temperature:.2f}"[:4],
            }
        elif len(messages) == 1:
            return {
                "system": hashlib.sha256("".encode()).hexdigest(),
                "user": hashlib.sha256((messages[0]["content"]).encode()).hexdigest(),
                "model": model,
                "temperature": f"{temperature:.2f}"[:4],
            }

    def close(self):
        self.fast_cache.close()

    def get_cache_key(self, messages, model, temperature, response_model):
        if not all([isinstance(msg["content"], str) for msg in messages]):
            normalized_messages = "".join([str(msg["content"]) for msg in messages])
        else:
            normalized_messages = "".join([msg["content"] for msg in messages])
        normalized_model = model
        normalized_temperature = f"{temperature:.2f}"[:4]
        normalized_response_model = (
            str(response_model.schema()) if response_model else ""
        )
        return hashlib.sha256(
            (
                normalized_messages
                + normalized_model
                + normalized_temperature
                + normalized_response_model
            ).encode()
        ).hexdigest()



import asyncio
import unittest
import time
import json
from apropos.src.bench.bigcodebench.backends.modal import (
    execute_code_remotely_modal_async,
    app,
)
from apropos.src.bench.bigcodebench.main import BigCodeBenchComplete_Benchmark


async def run_question(index: int):
    benchmark = BigCodeBenchComplete_Benchmark()
    question = benchmark.train[index]
    result = await execute_code_remotely_modal_async(
        question.information, question.information["answer"]
    )
    return result


# @app.local_entrypoint()
async def main():
    benchmark = BigCodeBenchComplete_Benchmark()

    async def process_question(i):
        t0 = time.time()
        try:
            result = await run_question(i)
        except Exception as e:
            print(f"Index {i} hard failed with error: {e}")
            return None
        elapsed_time = time.time() - t0
        if result[0] and result[1]["errors"] == 0:
            # print(f"Index {i} completed successfully")
            pass
        else:
            print(f"Index {i} failed")
        return elapsed_time

    tasks = [
        process_question(i + 1) for i in range(0, 10)
    ]  # range(0, len(benchmark.train))
    times_taken = await asyncio.gather(*tasks)

    print(f"Average time taken: {sum(times_taken) / len(times_taken)}")
    import matplotlib.pyplot as plt

    plt.figure(figsize=(10, 6))
    plt.hist(times_taken, bins=20, edgecolor="black")
    plt.title("Distribution of Time Taken for BigCodeBench Questions")
    plt.xlabel("Time (seconds)")
    plt.ylabel("Frequency")
    plt.savefig("time_distribution.png")
    plt.close()

    print(f"Distribution plot saved as 'time_distribution.png'")
    times_taken_by_index = {i: t for i, t in enumerate(times_taken)}
    sorted_times = sorted(
        times_taken_by_index.items(), key=lambda x: x[1], reverse=True
    )
    top_5_longest = sorted_times[:5]
    for index, time_taken in top_5_longest:
        print(f"Index {index}: {time_taken:.2f} seconds")


if __name__ == "__main__":
    asyncio.run(main())
    # ModuleNotFoundError: No module named 'pytz'
